import sqlalchemy as sql
import pandas as pd

SCHEMA_VERSION = 0.1
metadata = sql.MetaData()

kgenealogic = sql.Table(
    "kgenealogic",
    metadata,
    sql.Column("key", sql.String, nullable=False, primary_key=True),
    sql.Column("value", sql.String, nullable=False),
)

kit = sql.Table(
    "kit",
    metadata,
    sql.Column("id", sql.Integer, nullable=False, primary_key=True),
    sql.Column("kitid", sql.String, nullable=False, unique=True, sqlite_on_conflict_unique='IGNORE'),
    sql.Column("name", sql.String),
    sql.Column("email", sql.String),
    sql.Column("sex", sql.String),
)

source = sql.Table(
    "source",
    metadata,
    sql.Column("kit", sql.Integer, sql.ForeignKey("kit.id"), nullable=False, primary_key=True),
    sql.Column("has_segments", sql.Boolean, nullable=False, default=False),
    sql.Column("has_triangles", sql.Boolean, nullable=False, default=False),
    sql.Column("has_negative", sql.Boolean, nullable=False, default=False),
)

segment = sql.Table(
    "segment",
    metadata,
    sql.Column("id", sql.Integer, nullable=False, primary_key=True),
    sql.Column("chromosome", sql.String, nullable=False, index=True),
    sql.Column("start", sql.Integer, nullable=False),
    sql.Column("end", sql.Integer, nullable=False),
    sql.Column("length", sql.Float, index=True),
    sql.Column("generated", sql.Boolean, nullable=False),
    sql.UniqueConstraint("chromosome", "start", "end", sqlite_on_conflict='IGNORE'),
)

match = sql.Table(
    "match",
    metadata,
    sql.Column("segment", sql.Integer, sql.ForeignKey("segment.id"), nullable=False, index=True),
    sql.Column("kit1", sql.Integer, sql.ForeignKey("kit.id"), nullable=False, index=True),
    sql.Column("kit2", sql.Integer, sql.ForeignKey("kit.id"), nullable=False, index=True),
    sql.UniqueConstraint("segment", "kit1", "kit2", sqlite_on_conflict='IGNORE'),
)

triangle = sql.Table(
    "triangle",
    metadata,
    sql.Column("segment", sql.Integer, sql.ForeignKey("segment.id"), nullable=False, index=True),
    sql.Column("kit1", sql.Integer, sql.ForeignKey("kit.id"), nullable=False, index=True),
    sql.Column("kit2", sql.Integer, sql.ForeignKey("kit.id"), nullable=False, index=True),
    sql.Column("kit3", sql.Integer, sql.ForeignKey("kit.id"), nullable=False, index=True),
    sql.UniqueConstraint("segment", "kit1", "kit2", "kit3", sqlite_on_conflict='IGNORE'),
)

overlap = sql.Table(
    "overlap",
    metadata,
    sql.Column("id", sql.Integer, nullable=False, primary_key=True),
    sql.Column("source", sql.Integer, sql.ForeignKey("source.kit"), nullable=False, index=True),
    sql.Column("target1", sql.Integer, sql.ForeignKey("kit.id"), nullable=False, index=True),
    sql.Column("target2", sql.Integer, sql.ForeignKey("kit.id"), nullable=False, index=True),
    sql.Column("segment", sql.Integer, sql.ForeignKey("segment.id"), nullable=False, index=True),
    sql.UniqueConstraint("source", "target1", "target2", "segment", sqlite_on_conflict='IGNORE'),
)

negative = sql.Table(
    "negative",
    metadata,
    sql.Column("overlap", sql.Integer, sql.ForeignKey("overlap.id"), nullable=False, index=True),
    sql.Column("neg_segment", sql.Integer, sql.ForeignKey("segment.id"), nullable=False, index=True),
    sql.UniqueConstraint("overlap", "neg_segment", sqlite_on_conflict='IGNORE'),
)

def initialize(engine):
    metadata.drop_all(engine)
    metadata.create_all(engine)
    kg_meta = pd.DataFrame([
        dict(key='schema_version', value=str(SCHEMA_VERSION)),
        dict(key='cache_valid', value=str('N')),
    ])
    with engine.connect() as conn:
        kg_meta.to_sql("kgenealogic", conn, if_exists="append", index=False)
        conn.commit()



"""Base package for kgenealogic"""

__app_name__ = "kgenealogic"
__version__ = "0.1.0"

from kgenealogic.schema import initialize
from kgenealogic.data import import_matches, import_triangles
from kgenealogic.cache import is_cache_valid, build_cache
from kgenealogic.cluster import cluster_data



import pandas as pd
import sqlalchemy as sql
import importlib.resources as resources

from kgenealogic.schema import *
import kgenealogic as kg

GENETMAP_PATH='genetmap.csv'
GENETMAP_DTYPE={'chromosome': str, 'position': int, 'cm': float}

def invalidate_cache(engine):
    stmt = (
        sql.update(kgenealogic)
        .where(kgenealogic.c.key=='cache_valid')
        .values(value='N')
    )
    label_sources = sql.update(source).values(has_negative=False)
    with engine.connect() as conn:
        conn.execute(stmt)
        conn.execute(label_sources)
        conn.commit()

def is_cache_valid(engine):
    stmt = (
        sql.select(kgenealogic.c.value=='Y')
        .where(kgenealogic.c.key=='cache_valid')
    )
    with engine.connect() as conn:
        result = conn.execute(stmt)
    return result.scalar_one()

_CACHE_TABLES = [negative, overlap]
def clear_cache(engine):
    invalidate_cache(engine)

    for t in _CACHE_TABLES:
        t.drop(engine)
        t.create(engine)

    drop_gen_segs = sql.delete(segment).where(segment.c.generated)
    with engine.connect() as conn:
        conn.execute(drop_gen_segs)
        conn.commit()

def build_cache(engine):
    from kgenealogic.data import as_internal_segment
    clear_cache(engine)

    valid_neg_targets = (
        sql.select(
            source.c.kit.label("source"),
            triangle.c.kit2.label("target"),
        )
        .join_from(source, triangle, source.c.kit==triangle.c.kit1)
        .where(source.c.has_triangles)
        .distinct()
        .cte()
    )
    t1, t2 = valid_neg_targets.alias(), valid_neg_targets.alias()
    m1, m2 = match.alias(), match.alias()
    s1, s2 = segment.alias(), segment.alias()
    select_overlap = (
        sql.select(
            t1.c.source,
            t1.c.target.label("target1"),
            t2.c.target.label("target2"),
            s1.c.chromosome,
            sql.func.max(s1.c.start, s2.c.start).label("start"),
            sql.func.min(s1.c.end, s2.c.end).label("end"),
        )
        .join_from(t1, t2, sql.and_(
            t1.c.source==t2.c.source,
            t1.c.target!=t2.c.target,
        ))
        .join(m1, sql.and_(
            t1.c.source==m1.c.kit1,
            t1.c.target==m1.c.kit2,
        ))
        .join(m2, sql.and_(
            t2.c.source==m2.c.kit1,
            t2.c.target==m2.c.kit2,
        ))
        .join(s1, m1.c.segment==s1.c.id)
        .join(s2, sql.and_(
            m2.c.segment==s2.c.id,
            s1.c.chromosome==s2.c.chromosome,
            s1.c.start < s2.c.end,
            s2.c.start < s1.c.end,
        ))
    )
    with engine.connect() as conn:
        match_overlap = pd.read_sql(select_overlap, conn)
    match_overlap = as_internal_segment(engine, match_overlap, generated=True)
    match_overlap = match_overlap[["source", "target1", "target2", "segment"]]
    with engine.connect() as conn:
        conn.execute(overlap.insert(), match_overlap.to_dict(orient="records"))
        conn.commit()
        
    select_neg_overlap = (
        sql.select(
            overlap.c.id.label("overlap"),
            s1.c.chromosome,
            s1.c.start,
            s1.c.end,
            s2.c.start.label("tri_start"),
            s2.c.end.label("tri_end"),
        )
        .join_from(overlap, s1, overlap.c.segment==s1.c.id)
        .join(triangle, sql.and_(
            overlap.c.source==triangle.c.kit1,
            overlap.c.target1==triangle.c.kit2,
            overlap.c.target2==triangle.c.kit3,
        ), isouter=True)
        .join(s2, triangle.c.segment==s2.c.id)
        .where(sql.or_(
            s2.c.id.is_(None),
            sql.and_(
                s2.c.chromosome==s1.c.chromosome,
                s2.c.start <= s1.c.end,
                s2.c.end >= s1.c.start,
            ),
        ))
    )
    with engine.connect() as conn:
        neg_overlap = pd.read_sql(select_neg_overlap, conn)

    neg_base = (
        neg_overlap[neg_overlap.tri_start.isnull()]
        .drop(columns=["tri_start", "tri_end"])
        .drop_duplicates()
    )
    neg_remain = []
    neg_group_cols = ["overlap", "chromosome"]
    for _, tri_data in neg_overlap.dropna().groupby(neg_group_cols):
        id_vals = tri_data[neg_group_cols].iloc[0].to_dict()
        start, end = tri_data[['start', 'end']].iloc[0]
        # negative triangulations are the raw triangulation (start to end),
        # excluding all the positive parts (tri_start to tri_end)
        for _, row in tri_data.sort_values('tri_start').iterrows():
            if row.tri_start > start:
                neg_remain.append(dict(
                    start=start,
                    end=int(row.tri_start),
                    **id_vals))
            start = int(row.tri_end)
        if end > start:
             neg_remain.append(dict(
                start=start,
                end=end,
                **id_vals))       
    neg_tri = pd.concat([neg_base, pd.DataFrame(neg_remain)], ignore_index=True).drop_duplicates()
    neg_tri = as_internal_segment(engine, neg_tri, generated=True).drop_duplicates()
    neg_tri = neg_tri[["overlap", "segment"]].rename(columns=dict(segment="neg_segment"))
    with engine.connect() as conn:
        conn.execute(negative.insert(), neg_tri.to_dict(orient="records"))
        conn.commit()

    label_sources = sql.update(source).values(has_negative=True)
    with engine.connect() as conn:
        neg_tri.to_sql("negative", conn, if_exists="append", index=False, chunksize=25000)
        conn.execute(label_sources)
        conn.commit()

    # compute cm lengths for all segments
    null_len_segs_query = sql.select(segment).where(segment.c.length.is_(None))
    with engine.connect() as conn:
        null_len_segs = pd.read_sql(null_len_segs_query, conn)

    genetmap = pd.read_csv(resources.files(kg).joinpath(GENETMAP_PATH), dtype=GENETMAP_DTYPE)
    genetmap = genetmap.sort_values("position")

    seg_len_start = pd.merge_asof(
        null_len_segs.sort_values("start"),
        genetmap,
        left_on="start",
        right_on="position",
        by="chromosome",
        direction="nearest"
    ).set_index("id").cm

    seg_len_end = pd.merge_asof(
        null_len_segs.sort_values("end"),
        genetmap,
        left_on="end",
        right_on="position",
        by="chromosome",
        direction="nearest"
    ).set_index("id").cm

    null_len_segs['length'] = null_len_segs.id.map(seg_len_end-seg_len_start)
    null_len_segs = null_len_segs[['id', 'length']].rename(columns=dict(id="b_id"))

    seg_len_update = (
        segment.update()
        .where(segment.c.id==sql.bindparam("b_id"))
        .values(length=sql.bindparam("length"))
    )
    with engine.connect() as conn:
        conn.execute(seg_len_update, null_len_segs.to_dict(orient="records"))
        conn.commit()

    # all done, set cache as valid
    validate_cache = (
        sql.update(kgenealogic)
        .where(kgenealogic.c.key=='cache_valid')
        .values(value='Y')
    )
    with engine.connect() as conn:
        conn.execute(validate_cache)
        conn.commit()



import sqlalchemy as sql
from sqlalchemy.dialects.sqlite import insert as sqlite_insert
import pandas as pd

import kgenealogic.schema as kg

def add_sources(engine, sources, has_segments=None, has_triangles=None):
    traits = {}
    if has_segments is not None:
        traits['has_segments'] = has_segments
    if has_triangles is not None:
        traits['has_triangles'] = has_triangles
    stmt = sqlite_insert(kg.source).values(kit=sql.bindparam('kit'), **traits)
    if traits:
        stmt = stmt.on_conflict_do_update(index_elements=['kit'], set_=traits)
    else:
        stmt = stmt.on_conflict_do_nothing(index_elements=['kit'])

    with engine.connect() as conn:
        conn.execute(stmt, [dict(kit=x) for x in sources])
        conn.commit()

def as_internal_kitid(engine, data, kitid_fields):
    kit_ids = pd.concat([data[c] for c in kitid_fields]).drop_duplicates()
    insert_kit_id = kg.kit.insert().values(kitid=sql.bindparam('kit'))
    select_kit_id = sql.select(kg.kit.c["id", "kitid"])
    with engine.connect() as conn:
        conn.execute(insert_kit_id, [dict(kit=x) for x in kit_ids])
        conn.commit()
        kit_ids = pd.read_sql(select_kit_id, conn)

    kit_ids = kit_ids.set_index('kitid')['id']
    for c in kitid_fields:
        data = data.assign(**{c: data[c].map(kit_ids)})
    return data

def as_internal_segment(engine, data, generated=False):
    segments = (
        data[["chromosome", "start", "end"]]
        .groupby(["chromosome", "start", "end"])
        .head(1)
        .assign(generated=generated)
    )
    select_segment_ids = sql.select(kg.segment.c.id.label("segment"), kg.segment.c["chromosome", "start", "end"])
    with engine.connect() as conn:
        segments.to_sql("segment", conn, if_exists="append", index=False)
        conn.commit()
        segment_ids = pd.read_sql(select_segment_ids, conn)

    data = data.merge(segment_ids, on=["chromosome", "start", "end"])
    return data

def update_kit_data(engine, kit_data):
    insert_kit_data = (
        kg.kit.update()
        .where(kg.kit.c.id==sql.bindparam("kit"))
        .where(kg.kit.c.sex.is_(None))
        .values(
            name=sql.bindparam("name"),
            email=sql.bindparam("email"),
            sex=sql.bindparam("sex"),
        )
    )
    with engine.connect() as conn:
        conn.execute(insert_kit_data, kit_data.to_dict(orient='records'))
        conn.commit()

def import_matches(engine, matches):
    """ matches fields: kit1, kit2, chromosome, start, end, name, email, sex"""
    from .cache import invalidate_cache
    invalidate_cache(engine)

    matches = as_internal_kitid(engine, matches, ['kit1', 'kit2'])

    kit_data = (
        matches[['kit2', 'name', 'email', 'sex']]
        .groupby(['kit2'])
        .head(1)
        .drop_duplicates()
        .rename(columns=dict(kit2='kit'))
    )
    update_kit_data(engine, kit_data)

    sources = matches['kit1'].drop_duplicates()
    add_sources(engine, sources, has_segments=True)

    matches = as_internal_segment(engine, matches)

    matches = matches[['kit1', 'kit2', 'segment']]
    matches = pd.concat([matches, matches.rename(columns=dict(kit1='kit2', kit2='kit1'))],
                        ignore_index=True)
    with engine.connect() as conn:
        matches.to_sql("match", conn, if_exists="append", index=False)
        conn.commit()


def import_triangles(engine, triangles):
    from .cache import invalidate_cache
    invalidate_cache(engine)

    triangles = as_internal_kitid(engine, triangles, ['kit1', 'kit2', 'kit3'])

    kit_data = (
        pd.concat([
            triangles[['kit2', 'name2', 'email2']].rename(
                columns=dict(kit2='kit', name2='name', email2='email')
            ),
            triangles[['kit3', 'name3', 'email3']].rename(
                columns=dict(kit3='kit', name3='name', email3='email')
            ),
        ], ignore_index=True)
        .dropna()
        .drop_duplicates()
    )
    kit_data['sex'] = None
    update_kit_data(engine, kit_data)

    sources = triangles['kit1'].drop_duplicates()
    add_sources(engine, sources, has_triangles=True)

    triangles = as_internal_segment(engine, triangles)

    triangles = triangles[['kit1', 'kit2', 'kit3', 'segment']]
    triangles = pd.concat([
            triangles,
            triangles.rename(columns=dict(kit1='kit2', kit2='kit1')),
            triangles.rename(columns=dict(kit1='kit3', kit3='kit1')),
            triangles.rename(columns=dict(kit2='kit3', kit3='kit2')),
            triangles.rename(columns=dict(kit1='kit2', kit2='kit3', kit3='kit1')),
            triangles.rename(columns=dict(kit1='kit3', kit2='kit1', kit3='kit2')),
        ], ignore_index=True)
    with engine.connect() as conn:
        triangles.to_sql("triangle", conn, if_exists="append", index=False)
        conn.commit()


__all__ = [import_matches, import_triangles]



import numpy as np
import scipy as sp
import pandas as pd
import sqlalchemy as sql
from dataclasses import dataclass, field
import typing
import typer

from kgenealogic.schema import *

# factor for weighting pairwise matches in the graph; triangles have weight 1
PAIRWISE_FACTOR = 0.25

@dataclass
class Seed:
    kit: int
    floating: bool = field(default_factory=bool)
    negative: bool = field(default_factory=bool)

@dataclass
class SeedTree:
    """Data structure for organizing cluster seeds in tree"""
    ahnentafel: int
    values: list[Seed] = field(default_factory=list)
    branches: dict[str, typing.Any] = field(default_factory=dict)

def flatten(tree, values):
    for b in tree.branches.values():
        flatten(b, values)
    values.extend([v.kit for v in tree.values])
    return values

def cluster_data(engine, config):
    with engine.connect() as conn:
        kits = pd.read_sql(sql.select(kit.c['id', 'kitid']), conn)
    kits = kits.set_index('kitid').id

    seeds = set()
    exclude = set()
    for kitid_list, kit_set in ((config.seeds, seeds), (config.exclude, exclude)):
        for kitid in kitid_list:
            if kitid not in kits.index:
                typer.echo(f"Invalid kit id {kitid}", err=True)
                raise typer.Exit(code=1)
            else:
                kit_set.add(int(kits.loc[kitid]))

    if config.include:
        include = set()
        match_include_query = (
            sql.select(sql.distinct(match.c.kit2).label("kit"))
            .join_from(match, segment)
            .where(match.c.kit1==sql.bindparam("kit"))
            .where(segment.c.length > sql.bindparam("length"))
        )
        tri_include_query = (
            sql.select(sql.distinct(triangle.c.kit2).label("kit"))
            .join_from(triangle, segment)
            .where(triangle.c.kit1==sql.bindparam("kit"))
            .where(segment.c.length > sql.bindparam("length"))
        )
        with engine.connect() as conn:
            for k in config.include:
                kitid = k['id']
                if kitid not in kits.index:
                    typer.echo(f"Invliad kit id {kitid}", err=True)
                    raise typer.Exit(code=1)
                else:
                    include.add(int(kits.loc[kitid]))
                kitloc = int(kits.loc[kitid])
                if k.get('matches', None) is not None:
                    query_params = dict(kit=kitloc, length=k['matches'])
                    k_nbr = pd.read_sql(match_include_query, conn, params=query_params)
                    include.update([int(x) for x in k_nbr.kit])
                if k.get('triangles', None) is not None:
                    query_params = dict(kit=kitloc, length=k['triangles'])
                    k_nbr = pd.read_sql(tri_include_query, conn, params=query_params)
                    include.update([int(x) for x in k_nbr.kit])
        kits = kits[kits.isin(include|seeds)]

    kits = kits[~kits.index.isin(exclude)]

    graph_query = (
        sql.select(
            match.c.kit1,
            match.c.kit2,
            sql.func.sum(segment.c.length).label("weight"),
        )
        .join_from(match, segment, match.c.segment==segment.c.id)
        .where(segment.c.length >= config.min_length)
        .where(match.c.kit1 != match.c.kit2)
        .group_by(match.c['kit1', 'kit2'])
    )
    triangle_query = (
        sql.select(
            triangle.c.kit1,
            triangle.c.kit2,
            sql.func.sum(segment.c.length).label("weight"),
        )
        .join_from(triangle, segment, triangle.c.segment==segment.c.id)
        .where(segment.c.length >= config.min_length)
        .where(triangle.c.kit1 != triangle.c.kit2)
        .where(triangle.c.kit3.not_in(config.exclude))
        .group_by(triangle.c['kit1', 'kit2'])
    )
    trisource_query = (
        sql.select(source.c.kit)
        .where(source.c.has_triangles)
        .where(source.c.has_negative)
    )

    with engine.connect() as conn:
        graph = pd.read_sql(graph_query, conn)
        tri_graph = pd.read_sql(triangle_query, conn)
        trisource= set(int(x) for x in pd.read_sql(trisource_query, conn).kit)

    graph['weight'] = PAIRWISE_FACTOR*graph['weight'].astype(float)
    graph = (
        graph
        .set_index(['kit1', 'kit2'])
        .add(tri_graph.set_index(['kit1', 'kit2']), fill_value=0)
        .reset_index()
    )

    autox_query = (
        sql.select(sql.distinct(match.c.kit2).label('kit'))
        .join_from(match, segment, match.c.segment==segment.c.id)
        .where(segment.c.chromosome=='X')
        .where(segment.c.length >= config.min_length)
        .where(match.c.kit1==sql.bindparam('seed'))
    )
    def build_seed_tree(raw, ahnentafel):
        parsed = SeedTree(ahnentafel=ahnentafel)
        autox = set()
        for k in raw.get('kits', []):
            kitlocal = int(kits.loc[k['id']])
            floating = k['float']
            if floating is None:
                floating = kitlocal not in trisource
            if k['autox']:
                with engine.connect() as conn:
                    k_autox = pd.read_sql(autox_query, conn, params=dict(seed=kitlocal))
                k_autox = set(int(x) for x in k_autox.kit)
                k_autox.difference_update(exclude)
                k_autox.difference_update(seeds)
                seeds.update(k_autox)
                autox.update(k_autox)
            parsed.values.append(Seed(kit=kitlocal, floating=floating, negative=k['negative']))

        for a, short, long in ((1, 'M', 'maternal'), (0, 'P', 'paternal')):
            if long in raw:
                parsed.branches[short] = build_seed_tree(raw[long],a+(2*ahnentafel))

        if autox:
            if 'M' not in parsed.branches:
                parsed.branches['M'] = SeedTree(values=[],ahnentafel=1+(2*ahnentafel))
            maternal_values = parsed.branches['M'].values
            for x in autox:
                maternal_values.append(Seed(kit=x, floating=True))
        return parsed
    tree = build_seed_tree(config.tree, 1)

    source_tri = lambda source: get_triangles(engine, source, config.min_length)
    clusters = recursive_cluster(kits, tree, graph, source_tri)

    kitids = kits.reset_index().set_index('id').kitid
    clusters['kit'] = clusters.kit.map(kitids)

    fixed_cols = ['kit', 'ahnentafel']
    clusters = clusters[fixed_cols + [c for c in clusters.columns if c not in fixed_cols]]

    return clusters

def get_triangles(engine, source, min_length):
    neg_tri = (
        sql.select(
            overlap.c.target1.label("kit1"),
            overlap.c.target2.label("kit2"),
            sql.func.sum(-segment.c.length).label("weight"),
        )
        .join_from(negative, overlap)
        .join(segment, negative.c.neg_segment==segment.c.id)
        .where(overlap.c.source==source)
        .where(segment.c.length >= min_length)
        .group_by(overlap.c['target1', 'target2'])
    )

    with engine.connect() as conn:
        result = pd.read_sql(neg_tri, conn)
    return result

def recursive_cluster(kits, tree, graph, source_tri):
    # graph format: kit1/kit2/weight
    # initial value: all pairwise matches, suitably weighted relative to triangles that will be added?
    nonfloat = []
    tri_graph = graph
    for source in tree.values:
        if source.negative:
            tri_graph = tri_graph.set_index(['kit1', 'kit2']).add(
                source_tri(source.kit).set_index(['kit1', 'kit2']), fill_value=0
            ).reset_index()
        if not source.floating:
            nonfloat.append(source.kit)
        
    depth = int(np.log2(tree.ahnentafel))
    result = pd.DataFrame(dict(kit=kits))

    kits = kits[~kits.isin(nonfloat)]
    graph = graph[graph.kit1.isin(kits)&graph.kit2.isin(kits)]
    tri_graph = tri_graph[tri_graph.kit1.isin(kits)&tri_graph.kit2.isin(kits)]
    if tree.branches and (len(result)>0) and (len(tri_graph) > 0):
        flat_seeds = []
        for label, tree_branch in tree.branches.items():
            branch_seeds = flatten(tree_branch, [])
            branch_seeds = pd.DataFrame(dict(kit=branch_seeds))
            branch_seeds['label'] = label
            flat_seeds.append(branch_seeds)
        flat_seeds = pd.concat(flat_seeds, ignore_index=True)

        clusters = get_clusters(tri_graph, flat_seeds).set_index('kit')
        result['label'] = (
            result.kit.map(clusters.label)
            .fillna('')
            .astype(str)
        )
        #result.label.mask(result.kit.isin(flat_seeds.kit), flat_seeds.label, inplace=True)
        result['confidence'] = result.kit.map(clusters.confidence)
        result['ahnentafel'] = np.select(
            [result.label=='P', result.label=='M'],
            [2*tree.ahnentafel, 1+(2*tree.ahnentafel)], tree.ahnentafel
        )

        result_branches = []
        for label, kit_branch in result.groupby('label'):
            if label in tree.branches:
                tree_branch = tree.branches[label]
                branch = recursive_cluster(kit_branch.kit, tree_branch, graph, source_tri)
                branch[f"label{depth}"] = branch.kit.map(kit_branch.set_index('kit').label)
                branch[f"confidence{depth}"] = branch.kit.map(kit_branch.set_index('kit').confidence)
            else:
                branch = kit_branch.rename(columns=dict(
                    label=f"label{depth}",
                    confidence=f"confidence{depth}",
                ))
            result_branches.append(branch)
        result = pd.concat(result_branches, ignore_index=True)
    else:
        result['ahnentafel'] = tree.ahnentafel

    return result

def get_adjacency(edges, weights):
    vertex_to_id = (
        pd.concat(edges)
        .drop_duplicates()
        .rename('id')
        .reset_index(drop=True)
    )
    n = len(vertex_to_id)
    id_to_vertex = vertex_to_id.reset_index().set_index('id')['index']
    
    graph = sp.sparse.coo_array((weights, (edges[0].map(id_to_vertex), edges[1].map(id_to_vertex))), shape=(n,n)).toarray()
    return graph, vertex_to_id

def get_clusters(graph, seeds, max_rounds=None, fix_seeds=True):
    graph = graph.copy()
    labels = (
        graph
        .assign(weight=graph.weight.abs())
        .groupby('kit1')
        .weight
        .sum()
        .fillna(0)
        .reset_index()
        .rename(columns=dict(kit1='kit'))
    )
    labels['label'] = labels.kit.map(seeds.set_index('kit').label)
    labels = labels[labels.weight>0].set_index('kit').sort_index()
    
    max_rounds = max_rounds or 2*len(labels)
    i=0
    while True:
        graph['label'] = graph.kit2.map(labels.label)
        
        # TODO would be nice if seg_sum took confidence-weighted sum, but this is circular. does it converge?
        seg_sum = graph.groupby(['kit1', 'label']).weight.sum().unstack(level=-1).fillna(0)
        labels['paternal'] = seg_sum.P-seg_sum.M
        labels['confidence'] = labels['paternal'].abs()/labels['weight']
        available = np.any([
            labels.label.isnull(),
            (labels.label=='P')&(labels.paternal<0),
            (labels.label=='M')&(labels.paternal>0),
        ], axis=0)
        available = available & (labels.confidence > 0)
        if fix_seeds:
            available = available & (~labels.index.isin(seeds.kit))
        i+=1
        
        if (not np.any(available)) or (i > max_rounds):
            break

        next_kit = labels[available].iloc[labels.confidence[available].argmax()]
        next_label = 'P' if next_kit.paternal>0 else 'M'
        labels.loc[next_kit.name, 'label'] = next_label

    labels['label'] = labels.label.fillna('').astype(str)
    return labels.reset_index()[['kit', 'label', 'confidence']]



import pandas as pd
import strictyaml as yaml
from collections import defaultdict
from dataclasses import dataclass, field
import typing
import typer

GED_MATCH_COLS = {
    'PrimaryKit': 'kit1',
    'MatchedKit': 'kit2',
    'chr': 'chromosome',
    'B37Start': 'start',
    'B37End': 'end',
    'Segment cM': 'length',
    'MatchedName': 'name',
    'Matched Sex': 'sex',
    'MatchedEmail': 'email',
}
def is_ged_matches(path):
    try:
        csv = pd.read_csv(path, dtype=str)
    except:
        return False
    return set(GED_MATCH_COLS.keys()).issubset(csv.columns)

def read_ged_matches(path):
    dtype = defaultdict(lambda: str)
    dtype['B37Start'] = int
    dtype['B37End'] = int
    matches = pd.read_csv(path, dtype=dtype)
    return matches[GED_MATCH_COLS.keys()].rename(columns=GED_MATCH_COLS)

GED_TRIANGLE_COLS = {
    'Kit1 Number': 'kit2',
    'Kit1 Name': 'name2',
    'Kit1 Email': 'email2',
    'Kit2 Number': 'kit3',
    'Kit2 Name': 'name3',
    'Kit2 Email': 'email3',
    'Chr': 'chromosome',
    'B37 Start': 'start',
    'B37 End': 'end',
    'cM': 'length',
}
def is_ged_triangles(path):
    try:
        csv = pd.read_csv(path, dtype=str)
    except:
        return False
    return set(GED_TRIANGLE_COLS.keys()).issubset(csv.columns)

def read_ged_triangles(path, primary_kit):
    dtype = defaultdict(lambda: str)
    dtype['B37 Start'] = int
    dtype['B37 End'] = int
    tri = pd.read_csv(path, dtype=dtype)
    tri = tri[GED_TRIANGLE_COLS.keys()].rename(columns=GED_TRIANGLE_COLS)
    tri['kit1'] = primary_kit
    return tri

CLUSTER_KIT_SCHEMA = yaml.Str() | yaml.Map({
    'id': yaml.Str(),
    yaml.Optional('autox', default=False): yaml.Bool(),
    yaml.Optional('float', default=None, drop_if_none=False): yaml.EmptyNone() | yaml.Bool(),
    yaml.Optional('negative', default=None, drop_if_none=False): yaml.EmptyNone() | yaml.Bool(),
})
CLUSTER_INCLUDE_SCHEMA = yaml.Str() | yaml.Map({
    'id': yaml.Str(),
    yaml.Optional('matches', default=None, drop_if_none=False): yaml.EmptyNone() | yaml.Float(),
    yaml.Optional('triangles', default=None, drop_if_none=False): yaml.EmptyNone() | yaml.Float(),
})
CLUSTER_TREE_SCHEMA = yaml.Map({
    yaml.Optional('kits'): yaml.EmptyList() | yaml.Seq(CLUSTER_KIT_SCHEMA),
    yaml.Optional('maternal'): yaml.Any(),
    yaml.Optional('paternal'): yaml.Any(),
})
CLUSTER_CONFIG_SCHEMA = yaml.Map({
    yaml.Optional('include'): yaml.EmptyList() | yaml.Seq(CLUSTER_INCLUDE_SCHEMA),
    yaml.Optional('exclude'): yaml.EmptyList() | yaml.UniqueSeq(yaml.Str()),
    yaml.Optional('min_length'): yaml.Float(),
    'tree': CLUSTER_TREE_SCHEMA,
})

@dataclass
class ClusterConfig:
    """Data structure for storing parsed cluster configuration"""
    include: list[dict[str, typing.Any]]
    exclude: list[str]
    min_length: float
    tree: dict[str, typing.Any]
    seeds: set[str]

def validate_config_tree(config, seeds):
    expanded_kits = []
    for k in config.data.get('kits', []):
        if type(k)==str:
            expanded_kits.append(dict(id=k, autox=False, float=None, negative=False))
        else:
            expanded_kits.append(k)
    config['kits']=expanded_kits
    seeds.extend(expanded_kits)

    for branch in ('maternal', 'paternal'):
        if branch in config.data:
            config[branch].revalidate(CLUSTER_TREE_SCHEMA)
            validate_config_tree(config[branch], seeds)

def read_cluster_config(path):
    with open(path) as f:
        config_yaml = f.read()
    parsed = yaml.load(config_yaml, CLUSTER_CONFIG_SCHEMA)
    seeds = []
    validate_config_tree(parsed['tree'], seeds)
    exclude = parsed.data.get('exclude', [])

    include = []
    for k in parsed.data.get('include', []):
        if type(k)==str:
            include.append(dict(id=k, matches=None, triangles=None))
        else:
            include.append(k)

    unique_seeds = set()
    for s in seeds:
        if s['id'] in unique_seeds:
            typer.echo("duplicated seed {}".format(s['id']), err=True)
            raise typer.Exit(code=1)
        else:
            unique_seeds.add(s['id'])
    exclude_seeds = unique_seeds.intersection(exclude)
    if exclude_seeds:
        k = exclude_seeds.pop()
        typer.echo(f"excluded kit {k} is listed as seed", err=True)
        raise typer.Exit(code=1)

    result =  ClusterConfig(
        include=include,
        exclude=exclude,
        min_length=parsed.data.get('min_length', 7.),
        tree=parsed['tree'].data,
        seeds=unique_seeds,
    )
    return result

def write_clusters(clusters, path):
    clusters.to_csv(path, index=False)



"""CLI for kgenealogic"""


from pathlib import Path
from typing import Optional, List
from typing_extensions import Annotated
import sys

import sqlalchemy as sql
import typer

import kgenealogic as kg
import kgenealogic.files as kg_files

import warnings
warnings.filterwarnings("ignore",
    category=UserWarning,
    message="pandas only supports SQLAlchemy connectable"
)

app = typer.Typer()

DEFAULT_PROJECT = Path("kgenealogic.db")
DEFAULT_OUTPUT = Path("kg_results.csv")
PROJECT_OPTION = Annotated[
    Optional[Path],
    typer.Option(
        "--project",
        "-p",
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        resolve_path=True,
        help=f"Project file for storing processed genealogic data",
    ),
]

@app.command()
def init(
    project: PROJECT_OPTION = DEFAULT_PROJECT,
    force:  Annotated[bool, typer.Option(help="Force reinitialization of existing project file")] = False,
):
    """
    Initialize a new project file for storing processed genealogic data.

    If the project file already exists, --force clears and reinitializes the project.
    """
    if project.exists() and not force:
        typer.echo("Project already exists. Use --force to reinitialize.", err=True)
        raise typer.Abort()

    engine = sql.create_engine(f"sqlite:///{project}")
    kg.initialize(engine)

@app.command()
def add(
    files: List[Path],
    source:  Annotated[
        Optional[str],
        typer.Option(help="For GEDmatch triangulation files, the source kit number that is triangulated")
    ] = None,
    project: PROJECT_OPTION = DEFAULT_PROJECT,
):
    """
    Add data files to a project.

    Recognized file types:

    - GEDmatch segment match files, in CSV format with fields PrimaryKit, MatchedKit, chr,
    B37Start, B37End, Segment cM, MatchedName, Matched Sex, MatchedEmail

    - GEDmatch triangulation files, in CSV format with fields Kit1 Number, Kit1 Name, Kit1 Email,
    Kit2 Number, Kit2 Name, Kit2 Email, Chr, B37 Start, B37 End, cM

    For GEDmatch triangulation files, the source/primary kit number is not given in the data file
    and must be passed separately with the --source option
    """
    engine = sql.create_engine(f"sqlite:///{project}")
    for path in files:
        if kg_files.is_ged_matches(path):
            data = kg_files.read_ged_matches(path)
            kg.import_matches(engine, data)
        elif kg_files.is_ged_triangles(path):
            if not source:
                typer.echo("Source kit number required for GEDmatch triangulation files, aborting.",
                      err=True)
                raise typer.Exit(code=1)
            data = kg_files.read_ged_triangles(path, source)
            kg.import_triangles(engine, data)
        else:
            typer.echo(f"Unrecognized file type: {path}", err=True)

@app.command()
def build(
    project: PROJECT_OPTION = DEFAULT_PROJECT,
    force:  Annotated[bool, typer.Option(help="Force re-build, even if already built")] = False
):
    """
    Process all added data files in preparation for clustering.

    - Uses all segments with provided lengths in order to build a model of crossover probability, to
    impute cM lengths of derived segments

    - Finds all "negative" triangles, where two target kits have overlapping matching segments on a
    common source kit, but the segment does not appear as a triangulation for the three kits
    """
    engine = sql.create_engine(f"sqlite:///{project}")
    if not force:
        if kg.is_cache_valid(engine):
            typer.echo("Project is already built. Use --force to re-build", err=True)
            raise typer.Abort()

    kg.build_cache(engine)

@app.command()
def cluster(
    config: Annotated[Path, typer.Argument(
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
        help=f"The configuration file describing how to cluster kits",
        )],
    outfile: Annotated[Path, typer.Option(
        "--outfile",
        "-o",
        file_okay=True,
        dir_okay=False,
        writable=True,
        resolve_path=True,
        help=f"The destination for the output"
        )] = DEFAULT_OUTPUT,
    project: PROJECT_OPTION = DEFAULT_PROJECT,
):
    """
    Cluster kits to predict family tree structure.

    A config file in (strict) YAML format must be provided. An example of such a file is
    distributed with this package. The valid keys for this config file are:

    - include: Kit numbers to be included are specified as a list. Each list entry can be either
      (1) just the kit number to be included, as a string, or (2) a mapping indicating a kit number
      and some of its matches to be included. In the second format, the valid keys are

      -- id: the (base) kit number as a string (required)
      -- matches (optional, floating point): if this is given, it is interpreted as the minimum
      length in cM of a match so that if kit2 matches the base kit anywhere with this length, then
      kit2 will also be included
      -- triangles (optional, floating point): if this is given, it is interpreted as the minimum
      length in cM of a triangle so that if kit2 participates in a triangle the base kit anywhere
      with this length, with any third kit, then kit2 will also be included.

      Seeds are automatically included.

    - exclude: Kit numbers to be excluded are specified as a list. E.g., known duplicate kits can
      be specified here, or individuals that are known to not respect the desired tree structure.
      Optional, defaults to an empty list. Anything listed (directly or by reference) in the
      "include" list and also in the "exclude" list will be excluded.

    - min_length: Minimum segment length to be considered, in cM. Optional, defaults to 7cM

    - tree: A tree giving the desired family structure to be produced, along with any kits known to reside at
      each node of the of the tree. The more accurate seeds are available, the more reliable and
      useful the results will be. This root node of the tree is specified with the "tree" key. At
      each node of the tree, you can optionally specify any of:
      1. a "kits" key, associated with a list of kit numbers known to reside at that node of the
      tree. See below for additional options.
      2. a "paternal" key, formatted as a node of the tree, specifying the paternal branch
      2. a "maternal" key, formatted as a node of the tree, specifying the maternal branch
      The tree key is required

    Entries of each "kits" list of the tree can be specified in one of two ways:

    - just the kit number as a string
    - a mapping with the following keys:
      -- id: the kit number as a string (required)
      -- autox: whether to classify kits that match this kit on the X chromosome as necessarily
      maternal (optional, boolean). If the kit corresponds to a male child of the parents
      represented by the paternal and maternal branches of this node of the tree, it makes sense to
      use a value of true. Otherwise omit the key or set it to false (the default)
      -- negative: whether to use negative triangulations for this node, if available (option,
      boolean, default false)
      -- float: whether to keep the kit at this node of the tree, or attempt to float it into the
      branches (optional, boolean). If float is false, this seed will be kept at this node of the
      tree - this is desired when the kit is known to be a descendant of the parents represented by
      the maternal and paternal branches at this node. If float is true, this seed will be pushed
      out as far as possible into the branches - this is desired when the kit's relation to the
      root is partially but not completely known (e.g., the kit is some relative of the maternal
      grandfather, but nothing more is known). If float is not set, the default behavior is to
      treat it as false if triangulations are available, and otherwise treat it as true.

    For example:

    exclude: # these kits will be excluded from the clustering
      - T000001
      - T000002
    min_length: 8.5
    tree:
      kits:
        - # me (male)
          id: T000003
          autox: true
          float: false
          negative: true
        - # my sister's son
          id: T000004
          float: false
      paternal:
        kits:
          - T000004 # my father's grand-niece
        paternal:
          kits:
            - T000005 # a distant cousin on my paternal grandfather's side
        maternal:
          # I don't know any relatives of my paternal grandmother, but I'd like any that are
          # discovered to be present in the output
      maternal:
        kits:
          - T000006 # a relative of my mother, but not my father, with no other information

    This will produce an output file separating my maternal and paternal relatives, with the
    paternal relatives further split into paternal grandfather and paternal grandmother relatives.
    The output file will be a CSV file with several columns:
    - kit, specifying the kit number
    - ahnentafel, the ahnentafel number of the most distant relative of the root to which the kit
      is determiend to be related
    - label<n>, where n varies from 0 to depth of the tree minus 1. This is 'M' if the kit is put
      on the maternal side at that depth, and P if it is put on the paternal side. The list of
      these labels is equivalent to the value in the ahnentafel field
    - confidence<n>, a number between 0 and 1 indicating a rough confidence with which the kit is
      given label<n>. These numbers should not be interpreted as probabilities. A value of 1
      indicates that all of the kit's matches have labels consistent with the kit's label at this
      depth.
    """

    engine = sql.create_engine(f"sqlite:///{project}")
    if not kg.is_cache_valid(engine):
        typer.echo("Project is not yet built. First run kgenealogic build", err=True)
        raise typer.Abort()
    config = kg_files.read_cluster_config(config)
    clusters = kg.cluster_data(engine, config)
    kg_files.write_clusters(clusters, outfile)

